import abc


class BaseRecord(abc.ABC):
    CREATED = 'created'
    UPDATED = 'updated'
    DELETED = 'deleted'
    UNKNOWN = 'unknown'

    def __init__(self, record):
        self.valid = True
        self._body = None
        self._record = record
        self._attributes = {}

    @property
    @abc.abstractmethod
    def body(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def operation(self):
        raise NotImplementedError



import inspect


def requirements(**kwargs):
    def decorator_func(func):

        def run_before(request, response):
            if kwargs.get('before') and callable(kwargs['before']):
                kwargs['before'](request, response, kwargs)

        def run_after(request, response):
            if kwargs.get('after') and callable(kwargs['after']):
                kwargs['after'](request, response, kwargs)

        def run_method(request, response):
            run_before(request, response)
            if not response.has_errors and kwargs.get('data_class') and inspect.isclass(kwargs['data_class']):
                data_class = kwargs['data_class'](request=request)
                func(data_class, response)
            elif not response.has_errors:
                func(request, response)
            if not response.has_errors:
                run_after(request, response)
            return response

        run_method.requirements = kwargs
        return run_method

    return decorator_func



import inspect

from acai_aws.common import logger
from acai_aws.common.records.exception import EventException
from acai_aws.common.records.event import Event as CommonEvent
from acai_aws.documentdb.event import Event as DocumentDBEvent
from acai_aws.dynamodb.event import Event as DynamoDBEvent
from acai_aws.firehose.event import Event as FirehoseEvent
from acai_aws.kinesis.event import Event as KinesisEvent
from acai_aws.msk.event import Event as MSKEvent
from acai_aws.mq.event import Event as MQEvent
from acai_aws.s3.event import Event as S3Event
from acai_aws.sns.event import Event as SNSEvent
from acai_aws.sqs.event import Event as SQSEvent


def requirements(**kwargs):

    def __find_event_source(event):
        if event.get('eventSource'):
            return event['eventSource']
        if event.get('deliveryStreamArn'):
            return event['deliveryStreamArn']
        if event.get('Records') and event['Records'][0].get('eventSource'):
            return event['Records'][0]['eventSource']
        if event.get('Records') and event['Records'][0].get('EventSource'):
            return event['Records'][0]['EventSource']
        raise EventException(message='no known record event source found')

    def __determine_event_type(event, context):
        event_clients = {
            'unknown': CommonEvent,
            'aws:docdb': DocumentDBEvent,
            'aws:dynamodb': DynamoDBEvent,
            'aws:lambda:events': FirehoseEvent,
            'aws:kafka': MSKEvent,
            'aws:mq': MQEvent,
            'aws:kinesis': KinesisEvent,
            'aws:s3': S3Event,
            'aws:sns': SNSEvent,
            'aws:sqs': SQSEvent
        }
        try:
            source = __find_event_source(event)
            return event_clients[source](event, context, **kwargs)
        except EventException as event_error:
            if kwargs.get('verbose'):
                logger.log(level='ERROR', log={'event': event, 'context': context, 'error': event_error})
            return event_clients['unknown'](event, context, **kwargs)

    def decorator_func(func):

        def run_before(records_event):
            if kwargs.get('before') and callable(kwargs['before']):
                kwargs['before'](records_event, kwargs)

        def run_after(records_event, result):
            if kwargs.get('after') and callable(kwargs['after']):
                kwargs['after'](records_event, result, kwargs)

        def run_function(event, context):
            records_event = __determine_event_type(event, context)
            run_before(records_event)
            if kwargs.get('data_class') and inspect.isclass(kwargs['data_class']):
                records_event.data_class = kwargs['data_class']
            result = func(records_event)
            run_after(records_event, result)
            return result

        return run_function

    return decorator_func



import unittest

from acai_aws.apigateway.endpoint import Endpoint
from acai_aws.apigateway.resolver.importer import ResolverImporter
from acai_aws.apigateway.response import Response


class EndpointTest(unittest.TestCase):
    handler_path = 'tests/mocks/apigateway/endpoint/directory_handlers'

    def test_endpoint_initializes(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        file_path = f'/{importer.get_handlers_path_abs()}/basic.py'
        import_path = 'tests.mocks.apigateway.endpoint.directory_handlers.basic'
        endpoint_module = importer.import_module_from_file(file_path, import_path)
        endpoint = Endpoint(endpoint_module, 'post')
        self.assertTrue(isinstance(endpoint, Endpoint))

    def test_endpoint_has_requirements(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        file_path = f'/{importer.get_handlers_path_abs()}/basic.py'
        import_path = 'tests.mocks.apigateway.endpoint.directory_handlers.basic'
        endpoint_module = importer.import_module_from_file(file_path, import_path)
        endpoint = Endpoint(endpoint_module, 'post')
        self.assertTrue(endpoint.has_requirements)

    def test_endpoint_has_no_requirements(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        file_path = f'/{importer.get_handlers_path_abs()}/basic.py'
        import_path = 'tests.mocks.apigateway.endpoint.directory_handlers.basic'
        endpoint_module = importer.import_module_from_file(file_path, import_path)
        endpoint = Endpoint(endpoint_module, 'patch')
        self.assertFalse(endpoint.has_requirements)

    def test_endpoint_requires_auth(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        file_path = f'/{importer.get_handlers_path_abs()}/basic.py'
        import_path = 'tests.mocks.apigateway.endpoint.directory_handlers.basic'
        endpoint_module = importer.import_module_from_file(file_path, import_path)
        endpoint = Endpoint(endpoint_module, 'get')
        self.assertTrue(endpoint.requires_auth)

    def test_endpoint_has_required_response(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        file_path = f'/{importer.get_handlers_path_abs()}/basic.py'
        import_path = 'tests.mocks.apigateway.endpoint.directory_handlers.basic'
        endpoint_module = importer.import_module_from_file(file_path, import_path)
        endpoint = Endpoint(endpoint_module, 'search')
        self.assertTrue(endpoint.has_required_response)

    def test_endpoint_has_required_route(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        file_path = f'/{importer.get_handlers_path_abs()}/basic.py'
        import_path = 'tests.mocks.apigateway.endpoint.directory_handlers.basic'
        endpoint_module = importer.import_module_from_file(file_path, import_path)
        endpoint = Endpoint(endpoint_module, 'delete')
        self.assertTrue(endpoint.has_required_route)
        self.assertEqual('/some/route/{id}', endpoint.required_route)

    def test_endpoint_supports_custom_requirements(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        file_path = f'/{importer.get_handlers_path_abs()}/basic.py'
        import_path = 'tests.mocks.apigateway.endpoint.directory_handlers.basic'
        endpoint_module = importer.import_module_from_file(file_path, import_path)
        endpoint = Endpoint(endpoint_module, 'put')
        self.assertTrue(endpoint.has_requirements)
        self.assertDictEqual({'custom_list': [1, 2, 3], 'custom_dict': {'key': 'value'}, 'custom_simple': 1}, endpoint.requirements)

    def test_endpoint_runs_empty_requirements(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        file_path = f'/{importer.get_handlers_path_abs()}/basic.py'
        import_path = 'tests.mocks.apigateway.endpoint.directory_handlers.basic'
        endpoint_module = importer.import_module_from_file(file_path, import_path)
        endpoint = Endpoint(endpoint_module, 'post')
        response = Response()
        result = endpoint.run({}, response)
        self.assertEqual('{"endpoint_directory_basic": "post"}', result.full['body'])

    def test_endpoint_runs_no_requirements(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        file_path = f'/{importer.get_handlers_path_abs()}/basic.py'
        import_path = 'tests.mocks.apigateway.endpoint.directory_handlers.basic'
        endpoint_module = importer.import_module_from_file(file_path, import_path)
        endpoint = Endpoint(endpoint_module, 'patch')
        response = Response()
        result = endpoint.run({}, response)
        self.assertEqual('{"endpoint_directory_basic": "patch"}', result.full['body'])



import json
import unittest

from acai_aws.common.records.record import Record


class CommonRecordsTest(unittest.TestCase):
    record = {'key': 'value'}

    def test_record_accepts_event(self):
        record = Record(self.record)
        self.assertDictEqual(record.body, self.record)

    def test_record_body_decodes_json(self):
        json_record = json.dumps(self.record)
        record = Record(json_record)
        self.assertDictEqual(record.body, self.record)

    def test_record_has_unknown_operation(self):
        record = Record(self.record)
        self.assertEqual(record.operation, record.UNKNOWN)



import json
import unittest

from acai_aws.documentdb.record import Record

from tests.mocks.documentdb import mock_event


class DocumentDBRecordTest(unittest.TestCase):
    basic_record = mock_event.get_basic()['events'][0]
    unknown_record = mock_event.get_basic('unknown')['events'][0]

    def test_record_accepts_event(self):
        record = Record(self.basic_record)
        event = self.basic_record.get('event')
        self.assertEqual(record.operation, record.CREATED)
        self.assertEqual(record.event_id, event.get('_id', {}).get('_data'))
        self.assertTrue('2023-02-16' in record.cluster_time)
        self.assertEqual(record.document_key, event.get('documentKey', {}).get('_id', {}).get('$oid'))
        self.assertEqual(record.full_document, event['fullDocument'])
        self.assertEqual(record.change_event, event['operationType'])
        self.assertEqual(record.db, event.get('ns', {}).get('db'))
        self.assertEqual(record.collection, event.get('ns', {}).get('coll'))

    def test_record_fails_operation_gracefully(self):
        record = Record(self.unknown_record)
        self.assertEqual(record.operation, record.UNKNOWN)

    def test_record_prints(self):
        try:
            record = Record(self.basic_record)
            print(record)
            self.assertTrue(True)
        except Exception as error:
            print(error)
            self.assertTrue(False)



import inspect

from acai_aws.generic.event import Event


def requirements(**kwargs):

    def decorator_func(func):

        def run_before(generic_event):
            if kwargs.get('before') and callable(kwargs['before']):
                kwargs['before'](generic_event, kwargs)

        def run_after(generic_event, result):
            if kwargs.get('after') and callable(kwargs['after']):
                kwargs['after'](generic_event, result, kwargs)

        def run_function(event, context):
            generic_event = Event(event, context)
            run_before(generic_event)
            if kwargs.get('data_class') and inspect.isclass(kwargs['data_class']):
                generic_event = kwargs['data_class'](event=generic_event)
            result = func(generic_event)
            run_after(generic_event, result)
            return result

        return run_function

    return decorator_func



from acai_aws.common import logger


def log(**settings):
    def decorator_func(func):
        captured = {'arguments': {}, 'result': None}

        def run_func(*args, **kwargs):
            captured['arguments']['args'] = list(args)
            captured['arguments']['kwargs'] = kwargs
            captured['result'] = func(*args, **kwargs)
            if settings.get('condition') and callable(settings['condition']):
                if settings['condition'](*args, **kwargs):
                    logger.log(level=settings.get('level', 'INFO'), log=captured)
            else:
                logger.log(level=settings.get('level', 'INFO'), log=captured)
            return captured['result']

        return run_func

    return decorator_func



import base64

from acai_aws.common.json_helper import JsonHelper
from acai_aws.base.record import BaseRecord


class Record(BaseRecord):

    @property
    def record_id(self):
        return self._record.get('recordId')

    @property
    def epoc_time_stamp(self):
        return self._record.get('approximateArrivalTimestamp')

    @property
    def shard_id(self):
        return self._record.get('kinesisRecordMetadata', {}).get('shardId')

    @property
    def partition_key(self):
        return self._record.get('kinesisRecordMetadata', {}).get('partitionKey')

    @property
    def time_stamp(self):
        return self._record.get('kinesisRecordMetadata', {}).get('approximateArrivalTimestamp')

    @property
    def sequence_number(self):
        return self._record.get('kinesisRecordMetadata', {}).get('sequenceNumber')

    @property
    def subsequence_number(self):
        return self._record.get('kinesisRecordMetadata', {}).get('subsequenceNumber')

    @property
    def data(self):
        b64_decoded = base64.b64decode(self._record.get('data')).decode('utf-8')
        return JsonHelper.decode(b64_decoded)

    @property
    def body(self):
        return self.data

    @property
    def operation(self):
        return self.UNKNOWN

    def __str__(self):
        return str({
            'record_id': self.record_id,
            'epoc_time_stamp': self.epoc_time_stamp,
            'shard_id': self.shard_id,
            'partition_key': self.partition_key,
            'time_stamp': self.time_stamp,
            'sequence_number': self.sequence_number,
            'subsequence_number': self.subsequence_number,
            'body': self.body,
            'operation': self.operation
        })



import datetime
import os
import traceback

import jsonpickle


class CommonLogger:

    def __init__(self):
        self.__json = jsonpickle
        self.__run_mode = os.getenv('RUN_MODE')
        self.__log_level = os.getenv('LOG_LEVEL', 'INFO')
        self.__stage_variable = os.getenv('LOG_LOCAL_STAGE_VARIABLE', 'STAGE')
        self.__stage_name = os.getenv('LOG_LOCAL_STAGE_NAME', 'local')
        self.__log_stage = os.getenv(self.__stage_variable)
        self.__json.set_encoder_options('simplejson', use_decimal=True)
        self.__json.set_preferred_backend('simplejson')
        self.log_levels = {
            'INFO': 1,
            'WARN': 2,
            'ERROR': 3
        }

    def log(self, **kwargs):
        default_log = {
            'level': kwargs.get('level', 'INFO'),
            'log': kwargs.get('log', {})
        }
        if not self.__run_mode == 'unittest' and self.__log_stage == self.__stage_name and self.__should_log(**default_log):
            self.__log_local(**default_log)
        elif not self.__run_mode == 'unittest' and self.__should_log(**default_log):
            self.__log(**default_log)

    def __log_local(self, **kwargs):
        stack = self.__clean_up_stack_trace(kwargs['log'])
        print(self.__json.encode({
            'level': kwargs['level'],
            'time': datetime.datetime.now(datetime.timezone.utc).isoformat(),
            'stack': stack,
            'log': kwargs['log']
        }, indent=4))

    def __log(self, **kwargs):
        stack = self.__clean_up_stack_trace(kwargs['log'])
        print(self.__json.encode({
            'level': kwargs['level'],
            'time': datetime.datetime.now(datetime.timezone.utc).isoformat(),
            'stack': stack,
            'log': kwargs['log']
        }))

    def __clean_up_stack_trace(self, log_object):
        if isinstance(log_object, Exception):
            return traceback.format_exc().splitlines()
        skip = False
        new_line_split = [stack.split('\n') for stack in traceback.format_stack()]
        stack_trace = []
        for new_split in new_line_split:
            for new_line in new_split:
                if 'acai/common/logger' in new_line:
                    skip = True
                if new_line and not skip:
                    stack_trace.append(new_line)
        return stack_trace

    def __should_log(self, **kwargs):
        current_log_level = self.log_levels[kwargs['level']]
        log_level_setting = self.log_levels[self.__log_level]
        return current_log_level >= log_level_setting



from acai_aws.apigateway.resolver.modes.base import BaseModeResolver
from acai_aws.apigateway.exception import ApiException


class MappingModeResolver(BaseModeResolver):

    def __init__(self, **kwargs):
        self.__handler_mapping = kwargs['handlers']
        kwargs['handlers'] = '/'
        super().__init__(**kwargs)

    def _get_file_and_import_path(self, request_path):
        path_list = self.__get_request_path_as_list(request_path)
        mapping_path_key = self.__get_matching_path_from_mapping(path_list)
        mapping_file_path = self.__try_get_mapping_file_path(mapping_path_key)
        file_path = self.__get_abs_file_path(mapping_file_path)
        import_path = self.importer.clean_path(mapping_file_path).replace(self.importer.file_separator, '.').replace('.py', '')
        return file_path, import_path

    def __try_get_mapping_file_path(self, mapping_path_key):
        mapping_file_path = self.__handler_mapping.get(mapping_path_key)
        if not mapping_file_path:
            mapping_file_path = self.__handler_mapping.get(f'{self.base_path}/{mapping_path_key}')
        if not mapping_file_path:
            mapping_file_path = self.__handler_mapping.get(f'/{mapping_path_key}')
        if not mapping_file_path:
            mapping_file_path = self.__handler_mapping.get(f'/{self.base_path}/{mapping_path_key}')
        return mapping_file_path

    def __get_abs_file_path(self, mapping_file_path):
        clean_path = self.importer.clean_path(mapping_file_path)
        base_dir = clean_path.split(self.importer.file_separator)[0]
        project_root = self.importer.get_project_root().split(base_dir)[0]
        join_list = ['', self.importer.clean_path(project_root), self.importer.clean_path(mapping_file_path)]
        return f'{self.importer.file_separator}'.join(join_list)

    def __get_request_path_as_list(self, request_path):
        base_path = request_path.replace(self.base_path, '')
        clean_base = self.importer.clean_path(base_path)
        return clean_base.split('/')

    def __get_matching_path_from_mapping(self, path_list):
        mapping_keys = [self.__clean_route_mapping(route).split('/') for route in list(self.__handler_mapping.keys())]
        matching_path = self.__match_mapping_with_path(mapping_keys, path_list)
        if len(matching_path) != len(path_list):
            raise ApiException(code=404, message='route not found')
        return f'{self.importer.file_separator}'.join(matching_path)

    def __clean_route_mapping(self, route):
        mapping_key = route.replace(self.base_path, '')
        return self.importer.clean_path(mapping_key)

    def __match_mapping_with_path(self, mapping_list, path_list):
        matching = []
        for mapping in mapping_list:
            matching = self.__match_mapping(mapping, path_list)
            if matching:
                break
        return matching

    def __match_mapping(self, route, path_list):
        matching = []
        if len(route) == len(path_list):
            for index, _ in enumerate(path_list):
                if path_list[index] == route[index] or route[index].startswith('{') and route[index].endswith('}'):
                    matching.append(route[index])
                else:
                    matching = []
        return matching



from acai_aws.common.json_helper import JsonHelper
from acai_aws.base.record import BaseRecord


class Record(BaseRecord):

    @property
    def body(self):
        return JsonHelper.decode(self._record)

    @property
    def operation(self):
        return self.UNKNOWN



from collections import OrderedDict


class ResolverCache:
    CACHE_ALL = 'all'
    CACHE_STATIC = 'static-only'
    CACHE_DYNAMIC = 'dynamic-only'

    def __init__(self, **kwargs):
        self.__cache = OrderedDict()
        self.__size = kwargs.get('cache_size', 128)
        self.__mode = kwargs.get('cache_mode', self.CACHE_ALL)

    def get(self, method_path):
        if method_path not in self.__cache:
            return None
        self.__cache.move_to_end(method_path)
        return self.__cache[method_path]

    def put(self, route_path, endpoint, is_dynamic_route=False):
        if self.__size is None:
            return
        if is_dynamic_route and self.__mode == self.CACHE_STATIC:
            return
        if not is_dynamic_route and self.__mode == self.CACHE_DYNAMIC:
            return
        self.__cache[route_path] = endpoint
        self.__cache.move_to_end(route_path)
        if self.__size != 0 and len(self.__cache) > self.__size:
            self.__cache.popitem(last=False)
